library fields.mirrors;

import 'dart:mirrors';

final _listClassMirror = reflectClass(List);
final _mapClassMirror = reflectClass(Map);
final _objectClassMirror = reflectClass(Object);
final _dynamicTypeMirror = currentMirrorSystem().dynamicType;

Type typeOf(TypeMirror tm) {
  if (tm == _dynamicTypeMirror || tm is! ClassMirror) {
    return dynamic;
  } else {
    return (tm as ClassMirror).reflectedType;
  }
}

bool isSimpleType(Type t) {
  switch(t) {
    case bool:
    case double:
    case int:
    case List:
    case Map:
    case Object:
    case num:
    case String:
    case Type:
      return true;
    default:
      if (dynamic == t) {
        return true;
      } else {
        ClassMirror cm = reflectClass(t);
        if (isAssignableFrom(_listClassMirror, cm) ||
            isAssignableFrom(_mapClassMirror, cm)) {
          return true;
        }
      }
      return false;
  }
}

/// Returns `true` of the given [type] and the given [other] represent the same
/// type, or if the [type] is in the inheritance hierarchy of [other], or if
/// [other] implements the interface of [type]; else returns `false`.
bool isAssignableFrom(ClassMirror type, ClassMirror other) {
  assert(type != null);
  assert(other != null);
  if (!other.isOriginalDeclaration) other = other.originalDeclaration;  
  if (other == type) return true;
  if (other.superclass == null) return type == _objectClassMirror;
  if (other.superclass == type) return true;
  if (other.superinterfaces.any((i) => i == type)) return true;
  return isAssignableFrom(type, other.superclass);
}
